package indexing;
import java.io.*;
import java.util.*;

/**
 * 
 * @author Johannes A. Stork
 *
 */
public class SortedIndex {
	/// Data structure of the index
	private Vector<IndexElement> index;
	
	public SortedIndex() {
		this.index = new Vector<IndexElement>();
	}
	
	public Vector<IndexElement> getIndex() {
		return (Vector<IndexElement>) this.index.clone();
	}
	
	public int getOccurences() {
		int occurences = 0;
		
		for(IndexElement ie : this.index) {
			occurences+= ie.occurences;
		}
		
		return occurences;
	}
	
	/**
	 * Creates a index from the content files
	 * @return Inverted index of the content of all files named in the file index file
	 */
	public void addContent(StringBuffer content, String contentFileName) {
		/// Create a index of only this file content
		Vector<IndexElement> fileIndex = this.indexContent(content, contentFileName);
			
		/// Merge the over all index with the index of the file
		this.index = this.mergeIndexes(this.index, fileIndex);
	}
	
	
	/**
	 * Load a index from a file
	 * @return The index stored in the backup file
	 */
	public void loadIndex(String fileName) {
		/// Prepare empty index	
		this.index = new  Vector<IndexElement>();
		
		try {
			/// Prepare file reading
			BufferedReader br = new BufferedReader(new FileReader(fileName));
			
			/// Look at each line
			while(br.ready()) {
				/// Line style: content occurences file1 file2 ...
				String line = br.readLine();
				
				StringTokenizer st = new StringTokenizer(line);
				
				/// Content token
				String content = st.nextToken();
				/// Occurences token
				int occurences = Integer.parseInt(st.nextToken());
				/// First file token
				String firstFileName = st.nextToken();
				
				/// Prepare index entry
				IndexElement ie = new IndexElement(content, firstFileName);
				ie.occurences = occurences;
				
				/// Read the other file names
				while(st.hasMoreTokens()) {
					ie.fileNames.add(st.nextToken());
				}
				
				/// Finally add the entry to the index
				this.index.add(ie);
			}
		}
		catch(Exception e) {
			e.printStackTrace(System.err);
		}
	}
	
	/**
	 * Write a index to the backup file
	 * @param index the index to store
	 */
	public void writeIndex(String fileName) {
		
		try{
			/// Prepare the backup file
			File file = new File(fileName);
			PrintWriter pw = new PrintWriter(file);
			
			/// Each index entry gets one line
			for(IndexElement ie : this.index) {
				/// First word in line is content
				pw.print(ie.content + " ");
				
				/// Next word is the number of occurences
				pw.print(ie.occurences + " ");
				
				/// Rest of the line are file names
				for(String s : ie.fileNames) {
					pw.print(s + " ");
				}
				
				/// Start next line
				pw.println();
			}
			
			pw.flush(); /// not needed
			pw.close();
		} 
		catch(Exception e) {
			e.printStackTrace(System.err);
		}
	}
	
	/**
	 * Merge two indexes
	 * @param indexA
	 * @param indexB
	 * @return the merged index sorted my content of entry
	 */
	private Vector<IndexElement> mergeIndexes(Vector<IndexElement> indexA, Vector<IndexElement> indexB) {
		/// Prepare empty index
		Vector<IndexElement> sortedIndex = new Vector<IndexElement>();		
		/// Add the one index
		sortedIndex.addAll(indexA);
		
		/// Add evere element from the other index one by one
		for(IndexElement ie : indexB) {
			this.insertIndexElementIntoIndex(ie, indexA);
		}		
		return indexA;
	}
	
	/**
	 * Insert an entry to an sorted index and keey the index consistent
	 * @param ie The entry
	 * @param i The index
	 */
	private void insertIndexElementIntoIndex(IndexElement ie, Vector<IndexElement> i) {
		int pos = Collections.binarySearch(i, ie);
		
		if(pos < 0) {
			/// No equal entry found
			i.add( -(pos + 1), ie);
		} else {
			/// Equal entry found
			IndexElement other = i.get(pos);
			
			for(String s : ie.fileNames)  {
				this.insertFileNameIntoFileNameList(s, other.fileNames);
			}
			
			other.occurences+= ie.occurences;
		}
	}
	
	/**
	 * Add a file name to an sorted file name list and keep the list consistent
	 * @param fileName The file name
	 * @param fileNameList The list
	 */
	private void insertFileNameIntoFileNameList(String fileName, Vector<String> fileNameList) {
		int pos = Collections.binarySearch(fileNameList, fileName);
		
		if(pos < 0) {
			/// No equal entry found
			fileNameList.add( -(pos + 1), fileName);
		} else {
			/// Equal entry found, do nothing in set mode
			
		}
	}
	
	/**
	 * Create a index of word sequence
	 * @param content The word sequence
	 * @param fileName The file name of the content file
	 * @return Index of the content given
	 */
	private Vector<IndexElement> indexContent(StringBuffer content, String fileName) {
		/// Prepare empty index
		Vector<IndexElement> fileIndex = new Vector<IndexElement>();
		
		StringTokenizer st = new StringTokenizer(content.toString());
		
		/// Look at each token
		while(st.hasMoreTokens()) {
			String s = st.nextToken();
			
			/// Apply a stupid filter (not needed)
			if((s.substring(0, 1)).matches("[a-zA-Z]") 
					&& s.matches("[a-z]*")) {
				/// The content word starts and ends with a letter and consists only of letters
				
				/// Create an entry for that word with one occrences
				IndexElement ie = new IndexElement(s.toLowerCase(), fileName);
		
				this.insertIndexElementIntoIndex(ie, fileIndex);
			}
		}
		
		//Collections.sort(fileIndex); /// Not needed
		return fileIndex;
	}
	
	public Vector<String> oneWordQuerie(String word) {
		/// Prepare return value
		Vector<String> hitList = new Vector<String>();
		
		/// Search for the word in the index
		int pos = Collections.binarySearch(this.index, new IndexElement(word, ""));
		
		if(pos >= 0) {
			/// Word is in the index
			IndexElement ie = this.index.get(pos);
			
			hitList = ie.fileNames;
		} 
		
		return hitList;
	}
	
	
	public String toString() {
		return "SortedIndex with " + this.index.size() + " elements";
	}
}
